import logging
from django.db.models.signals import pre_save, post_delete
from django.dispatch import receiver

from defusedxml import ElementTree as ET

from ..models import GARInstitution
from ..gar import get_gar_institution_list, delete_gar_subscription

logger = logging.getLogger(__name__)


@receiver(pre_save, sender=GARInstitution, dispatch_uid="get_id_ent")
def get_id_ent(sender, instance, **kwargs):
    institution_list = get_gar_institution_list()
    xml_data = institution_list.content

    root = ET.fromstring(xml_data)

    namespace = {"ns": "http://www.atosworldline.com/listEtablissement/v1.0/"}

    id_ent = None
    for etablissement in root.findall("ns:etablissement", namespace):
        uai = etablissement.find("ns:uai", namespace)
        if uai is not None and uai.text == instance.uai:
            # Found the correct UAI, now get the idENT
            id_ent_object = etablissement.find("ns:idENT", namespace)
            id_ent = id_ent_object.text if id_ent_object is not None else None
            continue

    if id_ent:
        instance.id_ent = id_ent
    else:
        logger.info(f"id ent doesn't exist for uai {instance.uai}")


@receiver(post_delete, sender=GARInstitution, dispatch_uid="delete_subscription_in_gar")
def delete_subscription_in_gar(sender, instance, **kwargs):
    delete_gar_subscription(instance.subscription_id)



import datetime
import requests
import time

from bs4 import BeautifulSoup
from datetime import datetime

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.forms import ModelForm, ValidationError

from .gar import (
    get_gar_certificate,
    get_gar_headers,
    get_gar_request_url,
)
from .models import GARInstitution

GAR_DISTRIBUTOR_ID = getattr(settings, "GAR_DISTRIBUTOR_ID", "")
GAR_RESOURCES_ID = getattr(settings, "GAR_RESOURCES_ID", "")
GAR_ORGANIZATION_NAME = getattr(settings, "GAR_ORGANIZATION_NAME", "")
GAR_SUBSCRIPTION_PREFIX = getattr(settings, "GAR_SUBSCRIPTION_PREFIX", "")
User = get_user_model()


class GARInstitutionForm(ModelForm):
    subscription_id = forms.CharField(
        widget=forms.TextInput(attrs={"readonly": "readonly"})
    )

    class Meta:
        model = GARInstitution
        fields = (
            "uai",
            "institution_name",
            "ends_at",
            "user",
            "project_code",
            "subscription_id",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["subscription_id"].initial = "{}{}".format(
            GAR_SUBSCRIPTION_PREFIX, int(time.time())
        )

        # Disable ends_at field on update form
        instance = getattr(self, "instance", None)
        if instance.pk:
            self.fields["ends_at"].disabled = True

    def clean(self):
        cleaned_data = super().clean()
        self._create_or_update_gar_subscription()
        return cleaned_data

    def clean_uai(self):
        return self.cleaned_data.get("uai").upper().strip()

    def _create_or_update_gar_subscription(self):
        """
        A GAR subscription needs to be created or
        updated when an action occurred in the Back Office.
        Creating a subscription is done via PUT method.
        Updating a subscription is done via POST method.
        """

        if not self.initial:
            response = self._get_response_from_gar(http_method="PUT")
            if response.status_code not in [201, 200]:
                raise ValidationError(response.text)
        else:
            response = self._get_response_from_gar(http_method="POST")
            if response.status_code != 200:
                raise ValidationError(response.text)

    def _get_response_from_gar(self, http_method):
        url = get_gar_request_url(self.cleaned_data["subscription_id"])
        cert = get_gar_certificate()
        headers = get_gar_headers()
        response = requests.request(
            http_method,
            url,
            data=self._get_gar_data_to_send(http_method=http_method),
            cert=cert,
            headers=headers,
        )

        if response.status_code == 409 and "existe deja" in response.text:
            response = self._get_response_from_gar(http_method="POST")

        return response

    def _get_gar_data_to_send(self, http_method=None):
        """
        This is the data that needs to be sent when creating a subscription (PUT)
        or when updating a subscription (POST).
        When updating a subscription, the uaiEtab child should be removed.
        """
        uai = self.clean_uai()
        project_code = self.cleaned_data.get("project_code")
        xml = """<?xml version="1.0" encoding="UTF-8"?>
        <abonnement xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
           <idAbonnement>{subscription_id}</idAbonnement>
           <idDistributeurCom>{distributor_id}</idDistributeurCom>
           <idRessource>{resources_id}</idRessource>
           <typeIdRessource>ark</typeIdRessource>
           <libelleRessource>{organization_name}</libelleRessource>
           <debutValidite>{start_date}</debutValidite>
           <finValidite>{end_date}T00:00:00</finValidite>
           <uaiEtab>{uai}</uaiEtab>
           <categorieAffectation>transferable</categorieAffectation>
           <typeAffectation>INDIV</typeAffectation>
           <nbLicenceGlobale>ILLIMITE</nbLicenceGlobale>
           <publicCible>ELEVE</publicCible>
           <publicCible>ENSEIGNANT</publicCible>
           <publicCible>DOCUMENTALISTE</publicCible>
           <publicCible>AUTRE PERSONNEL</publicCible>
           <codeProjetRessource>{project_code}</codeProjetRessource>
        </abonnement>""".format(
            subscription_id=self.cleaned_data["subscription_id"],
            distributor_id=GAR_DISTRIBUTOR_ID,
            resources_id=GAR_RESOURCES_ID,
            organization_name=GAR_ORGANIZATION_NAME,
            start_date=self._get_gar_start_date(http_method),
            end_date=self.cleaned_data.get("ends_at"),
            uai=uai,
            project_code=project_code,
        )

        if not project_code:
            xml = xml.replace(
                f"<codeProjetRessource>{project_code}</codeProjetRessource>", ""
            )

        if http_method == "POST":
            xml = xml.replace(f"<uaiEtab>{uai}</uaiEtab>", "")

        return xml

    def _get_gar_start_date(self, http_method):
        """
        The start date (debutValidite) is mandatory but cannot be changed when
        updating a subscription. If we try to update the subscription we have to make a
        GET request to retrieve the start date.
        """
        if http_method == "POST":
            uai = self.clean_uai()
            data = """<?xml version="1.0" encoding="UTF-8"?>
            <filtres xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
                  <filtre>
                        <filtreNom>idDistributeurCom</filtreNom>
                        <filtreValeur>{distributor_id}</filtreValeur>
                  </filtre> 
                  <filtre>
                        <filtreNom>uaiEtab</filtreNom>
                        <filtreValeur>{uai}</filtreValeur>
                  </filtre> 
            </filtres>""".format(
                distributor_id=GAR_DISTRIBUTOR_ID, uai=uai
            )
            cert = get_gar_certificate()
            headers = get_gar_headers()
            response = requests.request(
                "GET",
                "https://abonnement.gar.education.fr/abonnements",
                data=data,
                cert=cert,
                headers=headers,
            )
            soup = BeautifulSoup(response.text, "lxml")
            subscriptions = soup.findAll("abonnement")
            for subscription in subscriptions:
                if (
                    subscription.find("idabonnement").text
                    == self.cleaned_data["subscription_id"]
                ):
                    return subscription.find("debutvalidite").text

        return datetime.now().isoformat()



import csv

from django.urls import path
from django.contrib import admin, messages
from django.http import HttpResponse
from django.shortcuts import redirect, reverse
from django.utils.html import format_html

from .gar import get_gar_subscription, get_allocations
from .forms import GARInstitutionForm
from .models import GARInstitution


@admin.register(GARInstitution)
class GARInstitutionAdmin(admin.ModelAdmin):
    raw_id_fields = ("user",)
    list_display = ("institution_name", "user", "uai", "ends_at")
    list_select_related = ("user",)
    readonly_fields = ("id_ent", "gar_subscription_response", "get_allocations")
    ordering = ("institution_name",)
    search_fields = ("institution_name", "user__email", "uai", "project_code")
    list_filter = ["project_code"]
    form = GARInstitutionForm
    change_list_template = "admin/django_gar/change_list.html"

    @admin.display(description="Etat de l'abonnement dans le GAR")
    def gar_subscription_response(self, obj):
        if not obj.uai:
            return ""

        gar_subscription = get_gar_subscription(obj.uai, obj.subscription_id)

        if not gar_subscription:
            return (
                "L'abonnement n'existe pas dans le GAR. "
                "Vous pouvez le supprimer et en créer un nouveau."
            )

        response = ""
        for element in gar_subscription.find_all():
            response += f"{element.name} : {element.text}<br/>"

        return format_html(f"<code>{response}</code>")

    @admin.display(description="Etat des affectations")
    def get_allocations(self, obj):
        if not obj.uai:
            return ""

        response = get_allocations(subscription_id=obj.subscription_id)
        decoded_response = response.content.decode("utf-8")

        if response.status_code == 200 and decoded_response:
            lines = decoded_response.splitlines()
            reader = csv.reader(lines, delimiter=";")
            rows = list(reader)
            headers = rows[0]
            values = rows[1]
            allocations = ""
            for header, value in zip(headers, values):
                allocations += f"{header} : {value}<br/>"
        elif response.status_code == 200:
            allocations = "L'établissement n'a pas encore affecté la ressource.<br/>Les informations fournies par le webservice font l’objet d’un traitement asynchrone et sont par conséquent actualisées quotidiennement. Il peut être constaté une latence dans la prise en compte de changements en cas d’affectations / récupérations de licences au sein d’une même journée."
        else:
            allocations = decoded_response.get("message")

        return format_html(f"<code>{allocations}</code>")

    def get_urls(self):
        urlpatterns = super().get_urls()

        allocations_report_url = [
            path(
                "allocations-report/generate/",
                self.admin_site.admin_view(self.allocations_report),
                name="{app_label}_{model_name}_generate_allocations_report".format(
                    app_label=self.model._meta.app_label,
                    model_name=self.model._meta.model_name,
                ),
            )
        ]

        return allocations_report_url + urlpatterns

    def allocations_report(self, request):
        project_code = request.GET.get("project_code")

        if not project_code:
            messages.success(
                request,
                "Impossible de télécharger le rapport d’affectations, le code projet ressource est introuvable",
            )
            return redirect(
                reverse(
                    "admin:{}_{}_changelist".format(
                        self.model._meta.app_label, self.model._meta.model_name
                    )
                )
            )

        allocations_response = get_allocations(project_code=project_code)
        data = allocations_response.content.decode("utf-8")

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = (
            f'attachment; filename="rapport_affectations_{project_code}.csv"'
        )

        writer = csv.writer(response)
        rows = [line.split(";") for line in data.splitlines()]

        writer.writerows(rows)

        return response



import pytest
import requests

from django_gar.forms import GARInstitutionForm

pytestmark = pytest.mark.django_db


class TestGARInstitutionForm:
    def test_form_works_with_gar_when_creating_instance(
        self, form_data, mocker, response_from_gar
    ):
        # GIVEN
        mock_request = mocker.patch.object(
            GARInstitutionForm,
            "_get_response_from_gar",
            return_value=response_from_gar(201, "dummy response message"),
        )

        # WHEN
        form = GARInstitutionForm(data=form_data().data)

        # THEN
        assert mock_request.called_once()
        assert form.is_valid()

    def test_form_error_with_gar_when_creating_instance(
        self, form_data, mocker, response_from_gar
    ):
        # GIVEN
        mock_request = mocker.patch.object(
            GARInstitutionForm,
            "_get_response_from_gar",
            return_value=response_from_gar(400, "dummy error message"),
        )

        # WHEN
        form = GARInstitutionForm(data=form_data().data)

        # THEN
        assert mock_request.called_once()
        assert not form.is_valid()

    def test_form_works_with_gar_when_try_creating_instance_that_already_exists(
        self, form_data, mocker, response_from_gar
    ):
        # GIVEN
        mock_request = mocker.patch.object(
            requests,
            "request",
            side_effect=[
                response_from_gar(409, "Cette abonnement existe deja"),
                response_from_gar(200, "Hello"),
                response_from_gar(201, "OK"),
            ],
        )

        # WHEN
        form = GARInstitutionForm(data=form_data().data)
        form.save()

        # THEN
        assert mock_request.call_count == 3
        assert form.is_valid()

    def test_form_works_with_gar_when_updating_instance(
        self, form_data, mocker, response_from_gar, user
    ):
        # GIVEN
        institution = user.garinstitution
        data = form_data(garinstitution=institution).data
        institution.save()
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "dummy response message"),
        )

        # WHEN
        form = GARInstitutionForm(instance=institution, data=data)
        form.save()

        # THEN
        assert mock_request.called_once()
        assert form.is_valid()

    def test_form_error_with_gar_when_updating_instance(
        self, form_data, mocker, response_from_gar, user
    ):
        # GIVEN
        institution = user.garinstitution
        data = form_data(garinstitution=institution).data
        institution.save()
        error_message = "dummy error message"
        mock_request = mocker.patch.object(
            requests, "request", return_value=response_from_gar(400, error_message)
        )

        # WHEN
        form = GARInstitutionForm(instance=institution, data=data)

        # THEN
        assert mock_request.called_once()
        assert not form.is_valid()



import pytest
import requests

from django_gar.exceptions import DjangoGARException
from django_gar.gar import get_gar_subscription, get_allocations, GAR_ALLOCATIONS_URL

pytestmark = pytest.mark.django_db()


class TestGetGarSubscription:
    def test_with_subscription_in_response(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.uai = 0561622j
        user.garinstitution.subscription_id = "briefeco_1630592291"
        user.garinstitution.save()
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(200, xml_response.read()),
            )

        # WHEN
        response = get_gar_subscription(
            user.garinstitution.uai, user.garinstitution.subscription_id
        )

        # THEN
        assert mock_request.call_count == 1
        assert response

    def test_with_wrong_subscription_id(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.uai = 0561622j
        user.garinstitution.save()
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(200, xml_response.read()),
            )

        # WHEN
        response = get_gar_subscription(
            user.garinstitution.uai, user.garinstitution.subscription_id
        )

        # THEN
        assert mock_request.call_count == 1
        assert not response

    def test_with_status_code_not_200(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.subscription_id = "briefeco_1630592291"
        user.garinstitution.save()
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(404, xml_response.read()),
            )

        # WHEN / THEN
        with pytest.raises(DjangoGARException):
            get_gar_subscription(
                user.garinstitution.uai, user.garinstitution.subscription_id
            )

        assert mock_request.call_count == 1


class TestGetAllocations:
    def test_get_allocations_no_params(self):
        # GIVEN / WHEN
        with pytest.raises(DjangoGARException) as exc_info:
            get_allocations()

        # THEN
        assert exc_info.value.status_code == 400
        assert (
            exc_info.value.message
            == "At least one of subscription_id or project_code is mandatory"
        )

    def test_get_allocations_both_params(self):
        # GIVEN / WHEN
        with pytest.raises(DjangoGARException) as exc_info:
            get_allocations(subscription_id="123", project_code="ABC")

        # THEN
        assert exc_info.value.status_code == 400
        assert (
            exc_info.value.message
            == "Cannot set subscription_id and project_code at the same time"
        )

    def test_get_allocations_with_subscription_id(
        self, user, mock_get_allocations_response
    ):
        # WHEN
        response = get_allocations(subscription_id=user.garinstitution.subscription_id)

        # THEN
        mock_get_allocations_response.assert_called_once_with(
            "GET",
            f"{GAR_ALLOCATIONS_URL}?idAbonnement={user.garinstitution.subscription_id}",
            cert=("", ""),
        )
        assert response.status_code == 200

    def test_get_allocations_with_project_code(self, mock_get_allocations_response):
        # WHEN
        response = get_allocations(project_code="DUMMY")

        # THEN
        mock_get_allocations_response.assert_called_once_with(
            "GET", f"{GAR_ALLOCATIONS_URL}?codeProjetRessource=DUMMY", cert=("", "")
        )
        assert response.status_code == 200



from django.db import models

from django.contrib.auth import get_user_model

User = get_user_model()


class GARInstitution(models.Model):
    uai = models.CharField(
        "Unité Administrative Immatriculée", max_length=14, unique=True
    )
    institution_name = models.CharField("Nom de l'institution", max_length=255)
    id_ent = models.CharField("ID de l'ent", max_length=255, null=True)
    ends_at = models.DateField("Date de fin d'abonnement", null=True)
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    subscription_id = models.CharField("id abonnement", max_length=255, unique=True)
    project_code = models.CharField(
        "Code de projet ressources", max_length=50, null=True, blank=True
    )

    def __str__(self):
        return f"{self.institution_name} ({self.uai})"


class GARSession(models.Model):
    """Store GAR active session. This will help us to delete user sessions when the user log out from the GAR"""

    ticket = models.CharField("CAS ticket", max_length=255, unique=True)
    session_key = models.CharField("Django session key", max_length=255, unique=True)

    def __str__(self):
        return f"ticket: {self.ticket} - session_key: {self.session_key}"
